#!/usr/bin/perl 

#######################################################
#  This Script is a Daemon for Whamb Sharing Protocol
#
#    Written by Jean-Edouard Babin (Jeb@jeb.com.fr)
#
# The script require Perl 5.7.3 or latter and module:
# - MP3::Info
# - Encode
# You can install them with 'perl -MCPAN -e shell'
#
# You also need RendezVous from Apple, it's open source
# http://developer.apple.com/darwin/projects/rendezvous/
#
# Read the README file if you need more info
#
#######################################################

use MP3::Info;
use POSIX qw(strftime);
use IO::Socket;
use Net::hostent;
use Encode qw/encode decode/;
use HTML::Entities;
#use strict;

$port = 7603;        # daemon port
$recursive = 1;      # 0 or 1; if 1 the script will look for audio file recursively
$allinadir = 1;      # create a directory on / of your pl
$audiofileexpr = '([mM][pP]3|[oO][gG][gG])$';  # Expression that match an audio file

my $debug = "CLIENTREQ|PRINTPLRANGE|PRINTGETFILE";
## ADDFILE to see what file are added to the Allow Table for GET
## NEWCLIENT to see the client connection
## CLIENTREQ to see client request
## PRINTIDDB to see $iddb[n] after init
## PRINTPLRANGE to see the id range of each pl
## PRINTGETFILE to see the real file path of the GET when getting ID
## PRINTPLSEND to see what is send when getting the PL
## PRINTSIGPIPE to see sigpipe

############### don't edit after this line ############

$SIG{"PIPE"} = \&sigPIPE_handler;
## we handle SIGPIPE to don't stop the app
#  my $sigset = POSIX::SigSet->new();
#  my $action = POSIX::SigAction->new('sigPIPE_handler', $sigset, &POSIX::SA_NODEFER);
#  POSIX::sigaction(&POSIX::SIGPIPE, $action);
## This is ths POSIX way to catch SIGPIPE, i don't use it because i get some perl core dump with it on some linux

$n = $k = $getnext = 0;
$htmlescape = "&<>\"";

foreach $cmlarg (@ARGV) {
	($plname,$plpath) = split(/=/,$cmlarg);
	@tpath = split(//,$plpath);
#	$mtpath = $tpath[$#t-3] . $tpath[$#t-2] . $tpath[$#t-1] . $tpath[$#t] ;    # file extension
	if (-f "$plpath") { $pllist{$plname}{path} = $plpath; } 
	elsif (-d "$plpath") {
		if ($tpath[$#tpath] ne "/") { $plpath = "$plpath" . "/" }
		$pllist{$plname}{path} = $plpath;
	}
	else { print "WARNING: Playlist $plname: $plpath is not a file neither a directory, i will ignore it !\n"; }
}

## Here we have $pllist{NAME_OF_THE_PL}{path} = PATH_OF_THE_PL (file or directory)

foreach $plname (keys %pllist) {
	$pllist{$plname}{idmin}= $k ;
	print "Adding file for playlist: $plname: $pllist{$plname}{path}\n" if ($debug =~ /ADDFILE/);
	if (-d "$pllist{$plname}{path}") {
		if ($allinadir) {
			$pllist{$plname}{path} =~ /.*\/(.*)\/$/; $dir = "$1";
			encode_entities($dir, $htmlescape);
			$iddb[$k] = "		<array>
			<dict>
				<key>Name</key>
				<string>$dir</string>
				<key>Open</key>
				<true/>
				<key>Path</key>
				<string></string>
				<key>Sync</key>
				<true/>
			</dict>\n"; $k++;
		}
		&createiddb("$pllist{$plname}{path}");
		if ($allinadir) { $iddb[$k] = "		</array>\n"; $k++; }
	}
        elsif (-f "$pllist{$plname}{path}") {
		## open shouldn't fail, because a first check is done when building $pllist{$plname}{path};
		open(PLIST,"<$pllist{$plname}{path}") || die "Can't open $pllist{$plname}{path}: $!\n";
		while(<PLIST>) {
			if ($getnext) { 
				$getnext = 0;
				if ($_ =~ /<string>(.*)<\/string>$/) {
					$file = $1;
					print "Adding File: $file\n" if ($debug =~ /ADDFILE/);
					encode_entities($file, $htmlescape);
					$iddb[$k] = $file;
					$filedb{"'$file'"} = $k;
					$k++;
				}
			}
			if ($_ =~ /<key>Path<\/key>$/ ) { $getnext = 1 }
		}
		close(PLIST);
        }
	$pllist{$plname}{idmax}= $k-1 ;
}

## ok init is finished ;)

if ($debug =~ /PRINTIDDB/) {
	while($iddb[$n]) { print '$iddb[' . $n . ']=' . $iddb[$n] . "\n"; $n++ }	
}

if ($debug =~ /PRINTPLRANGE/) {
	foreach $plname (keys %pllist) {
		print "$plname range: from $pllist{$plname}{idmin} to $pllist{$plname}{idmax}\n";
	}
}


## now we start the server 

$server = IO::Socket::INET->new( Proto     => 'tcp',
                                  LocalPort => $port,
                                  Listen    => SOMAXCONN,
                                  Reuse     => 1);
die "can't setup server" unless $server;
print "<<<Server $0 listening on port $port>>>\n";

## ok now we wait a client =)

while ($client = $server->accept()) {
	print "There is a New Client\n" if ($debug =~ /NEWCLIENT/);
	while (<$client>) {
		print "$_" if ($debug =~ /CLIENTREQ/); 
		   if ($_ =~ /^WINK WHSP\/1\.0/)          { &helloyou }
		elsif ($_ =~ /^LIST WHSP\/1\.0/)          { &listavpl }
		elsif ($_ =~ /^PLAYLIST (.*) WHSP\/1\.0/) { &listthepl($1) }
		elsif ($_ =~ /^GET (.*) (\d+) WHSP\/1\.0/) {
			$temp = $1; $offset = $2;
			if ($temp =~ /^\d+$/) { $fileid = $temp; }
			else {
				$filename = $temp;
				$filename =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("U", hex($1))/eg;
				$fileid = $filedb{"'$filename'"};
				if (! $fileid) { close $client }
			}
			if ($iddb[$fileid]) {
				print "GET $iddb[$fileid]\n" if ($debug =~ /PRINTGETFILE/);
				&getfile($iddb[$fileid],$offset);
			}
			else { close $client }
		}
		elsif ($_ =~ /^\n/) { }
		else { close $client }
	}
}

sub helloyou {
print $client &header;
$login = getlogin || getpwuid($<);
print $client <<"EOF";  
        <key>build</key>
        <string>1A00</string>
        <key>playlists count</key>
        <integer>1</integer>
        <key>user</key>
        <string>$login</string>
        <key>version</key>
        <string>WSD 1.1</string>
EOF
print $client &footer;
close($client);
}


sub listavpl {
	print $client &header;
	print $client  "	<key>playlists</key>\n";
	print $client  "	<array>\n";
	foreach (keys %pllist) {
		print $client "		<string>$_</string>\n"
	}
	print $client  "	</array>\n";
	print $client &footer;
	close($client);
}

sub listthepl {
	$plname = "$_[0]";
	if (-d "$pllist{$plname}{path}") {
		print $client &header;
		print $client  "	<key>Version</key>\n";
		print $client  "	<string>1.1</string>\n";
		print $client  "	<key>Playlist</key>\n";
		print $client  "	<array>\n";
#		&readdir("$pllist{$plname}{path}");
		&readdir($pllist{$plname}{idmin} , $pllist{$plname}{idmax});
		print $client "	</array>\n";
		print $client &footer;
	}
	elsif (-f "$pllist{$plname}{path}") {
		## open shouldn't fail, because a first check is done when building $pllist{$plname}{path};
		open (PL, "$pllist{$plname}{path}") || die "Can't open $pllist{$plname}{path}: $!\n";
		print $client <PL>;
		close (PL);
	}
	close($client);
}

sub getfile {
	$a = "$_[0]"; 
	$offset = $_[1];
	open(MP3,"<$a") || die "Can't open $a: $!\n";
	seek(MP3,$offset,0);
	while((read(MP3, $buf, 1024)) && ($client->connected)) {
		print $client $buf;
	}
	close(MP3);
	close($client) if ($client->connected);
}

sub readdir {
	my $min = "$_[0]";
	my $max = "$_[1]";
	for ($i = $min; $i <= $max; $i++ ) {
		$file = $iddb[$i];

		if ($file =~ /array>/ ) {
			print $client $file;
			print $file if ($debug =~ /PRINTPLSEND/);
			next;
		}		

		$ftype = "MP3F" if ($file =~ /\.[mM][pP]3$/);
		$ftype = "OGGF" if ($file =~ /\.[Oo][gG]{2}$/);
##		$ftype = "MP3S" if ($file =~ /^http/);           ## there is no http is a folder -> not usefull

		$album = $artist = $genre = $time = $title = "";
		$bitrate = $sectime = 0;
		$year = -1;
		my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$file");
		$mtime = strftime("%Y-%m-%dT%H:%M:%SZ",localtime($mtime));

		if ($ftype eq "MP3F") {
			my $tag = get_mp3tag("$file");
			my $info = get_mp3info("$file");
			$album = $tag->{ALBUM};
			if ($album ne "") {
				encode_entities($album, $htmlescape);
				$album = encode("UTF-8", $album);
			}
			$artist = $tag->{ARTIST};
			if ($artist  ne "") {
				encode_entities($artist, $htmlescape); 
				$artist = encode("UTF-8", $artist);
			}
			$bitrate = $info->{BITRATE};
			if ($bitrate eq "") {
				$bitrate="-1";
			}
			$genre = $tag->{GENRE};
			if ($genre ne "") {
				encode_entities($genre, $htmlescape); 
				$genre = encode("UTF-8", $genre);
			}
			$time = $info->{TIME};
			if ($time eq "") {
				$time = "-1";
			}
			$info->{SECS} =~ /^(\d+)/;
			$sectime = $1;
			if ($sectime eq "") {
				$sectime = "-1";
			}
			$title = $tag->{TITLE};
			if ($title ne "") {
				encode_entities($title, $htmlescape); 
				$title = encode("UTF-8", $title);
			}
			if ($tag->{YEAR}) {
				$year = $tag->{YEAR};
				if ($year =~ /(\d+)/) { $year = "$1" }
				else { $year = "-1"; }
			}
		}

		if ($title eq "") {
			$file =~ /.*\/(.*)$/;
			$title = $1;
			encode_entities($title, $htmlescape);
			$title = encode("UTF-8", $title);
		}

print $client <<"EOF";
		<dict>
			<key>Album</key>
			<string>$album</string>
			<key>Artist</key>
			<string>$artist</string>
			<key>BitRate</key>
			<integer>$bitrate</integer>
			<key>Current</key>
			<real>-1.0</real>
			<key>File Size</key>
			<integer>$size</integer>
			<key>File Type</key>
			<string>$ftype</string>
			<key>Genre</key>
			<string>$genre</string>
			<key>Mod. Date</key>
			<date>$mtime</date>
			<key>Path</key>
			<string>$i</string>
			<key>Time</key>
			<string>$time</string>
			<key>Time#</key>
			<integer>$sectime</integer>
			<key>Title</key>
			<string>$title</string>
			<key>TrackCount</key>
			<integer>0</integer>
			<key>TrackNum</key>
			<integer>0</integer>
			<key>Year</key>
			<string>$year</string>
			<key>Year#</key>
			<integer>$year</integer>
		</dict>
EOF

print  <<"EOF" if ($debug =~ /PRINTPLSEND/);
                <dict>
                        <key>Album</key>
                        <string>$album</string>
                        <key>Artist</key>
                        <string>$artist</string>  
                        <key>BitRate</key>
                        <integer>$bitrate</integer>
                        <key>Current</key>
                        <real>-1.0</real>
                        <key>File Size</key>
                        <integer>$size</integer>
                        <key>File Type</key>
                        <string>$ftype</string>
                        <key>Genre</key>
                        <string>$genre</string>
                        <key>Mod. Date</key>
                        <date>$mtime</date>
                        <key>Path</key>
                        <string>$i</string>
                        <key>Time</key>
                        <string>$time</string>
                        <key>Time#</key>
                        <integer>$sectime</integer>   
                        <key>Title</key>
                        <string>$title</string>
                        <key>TrackCount</key>
                        <integer>0</integer>
                        <key>TrackNum</key>   
                        <integer>0</integer>
                        <key>Year</key>
                        <string>$year</string>
                        <key>Year#</key>
                        <integer>$year</integer>
                </dict>
EOF

	}
}


sub createiddb {
	my $dirpath = "$_[0]";
	foreach $file ( glob ( "'$dirpath*'" ) ) {
		&filecreateiddb("$file");
	}
	foreach $file ( glob ( "$dirpath*" ) ) {
		&filecreateiddb("$file");
	}
}

sub filecreateiddb {
	my $dirpath = "$_[0]";
		if ((-f $file) && ($file =~ /$audiofileexpr/)) {
			print "Adding File: $file\n" if ($debug =~ /ADDFILE/);
			$iddb[$k] = $file;
			$k++;
		}
		if ((-d $file) && ($recursive eq 1)) {
			$file =~ s/\r//; $file =~ s/\n//;
			$file =~ /.*\/(.*)$/; $dir = "$1";
			$dir =~ s/&/&amp;/;
#### work ?
			$dir = encode("UTF-8", $dir);
####
			print "Going into Dir: $file/\n" if ($debug =~ /ADDFILE/);
			encode_entities($dir, $htmlescape);
			$iddb[$k] = "		<array>
			<dict>
				<key>Name</key>
				<string>$dir</string>
				<key>Open</key>
				<true/>
				<key>Path</key>
				<string></string>
				<key>Sync</key>
				<true/>
			</dict>\n"; $k++;
			&createiddb("$file/");
			$iddb[$k] = "		</array>\n"; $k++;
		}
}


sub header {
return <<"EOF";
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>                  
EOF
}

sub footer {
return <<"EOF";  
</dict>
</plist>
EOF
}


sub sigPIPE_handler {
	print "Ignoring the sigpipe\n" if ($debug =~ /PRINTSIGPIPE/);
}
